#! /usr/bin/perl
######################################################################
# Copyright 1995, 1996 Ipsilon Networks, Inc. All rights reserved.
# 
# YOU MAY ONLY USE THE MATERIALS FOUND IN THIS FILE PURSUANT TO THE
# TERMS OF THE EVALUATION LICENSE AGREEMENT FOUND IN "/LICENSE".  ALL
# OTHER USES ARE EXPRESSLY PROHIBITED.
######################################################################

# Global associative arrays used (based on interface name):
#	iflcladdr, ifmsklen, iflink, ifremaddr, ifseen
#	ifispnp, ifpnpremote

open (FD, "/bin/hostname |");
chop($hostname=<FD>);
close(FD);


sub stop_mrouted {
        system("/usr/local/config/bin/udw machine:$hostname:process:mrouted");
}
sub stop_ipsrd {
        system("/usr/local/config/bin/udw machine:$hostname:process:ipsrd");
}
sub start_ipsrd {
        system("/usr/local/config/bin/udw \"machine:$hostname:process:ipsrd t\"");
}

sub handler {	# 1st argument is signal name
	local($sig) = @_;
	print "Received reconfiguration request\n" if $debug;
	# See the startup to see why some of these things are done
	&stop_mrouted;
	system "route delete 224.0.0.0 -netmask 0xf0000000";
	$need_mcast_route = 1;
	$need_mrouted = 0;
	&parse_addrconf;

	open(IFCONFIG, "/sbin/ifconfig -a |") || return;
	while (<IFCONFIG>) {
		chop;
		# Scan through each interface line looking for 
		# interface names we don't know about
		if (/^(.*):.*flags=.*<(.*)>/) {
			if (!&is_known_iface($1)) {
				print "delete interface $1\n" if $debug;
				$ifcmd{$1}= "";
				system "ifconfig $1 down";
				system "ifconfig $1 delete";
				sleep 1;
			}
		}
	}
	close(IFCONFIG);
}

sub is_known_iface {
	local($iface) = @_;

	# Return true if the given interface is in the config file last
	# parsed or is special cased.  At this point the only special
	# case interface is the loopback interface lo0.

	# Check for special cases first
	print "checking interface $iface against lo0\n" if $debug;
	if ($iface eq "lo0") {
		return 1;
	}

	# Now search list of parsed interfaces (iflcladdr will always
	# be filled in for every interface).
	@iface_keys = keys %iflcladdr;
	foreach $key (@iface_keys) {
		print "checking interface $iface against $key\n" if $debug;
		if ($iface eq $key) {
			return 1;
		}
	}

	# The interface is not known
	return 0;
}

sub get_remote {
	local($iface) = @_;

	open(IFACE, "/sbin/ifconfig $iface |") || return;
	while (<IFACE>) {
		chop;
		# Look for IP address of remote side (take first address)
		if (/^\s*inet\s+(\S+)\s+-->\s+(\S+)/) {
			return $2;
		}
	}
	close(IFACE);
	return "0.0.0.0";
}

sub start_mrouted {

	if ($need_mcast_route) {
		print "Need multicast route to start mrouted\n" if $debug;
		return;
	}
	if (! $need_mrouted) {
		print "skipping mrouted start effort\n" if $debug;
		return;
	}
	$need_mrouted = 0;

	return unless -e "/etc/mrouted.conf";

	&stop_mrouted;
	# start it up again
        system("/usr/local/config/bin/udw \"machine:$hostname:process:mrouted t\"");
};

sub configif {
	($iface,$isptp,$ismcast,$isloop) = @_;
	print "config addr for $iface\n" if $debug;

	if (!$iflcladdr{$iface}) {
		# If we have no configuration we can't do anything else
		print "No configuration information for $iface\n"
			if $debug;
		return;
	}

	$cmd = "/sbin/ifconfig $iface $iflcladdr{$1}";
	if ($ifmsklen{$iface}) {
		$cmd .= "/$ifmsklen{$iface}";
	}

	if ($isptp) {
		print "$iface is a P2P interface\n" if $debug;
		$need_mrouted = 1;
		if ($ifremaddr{$iface}) {
			$cmd .= " $ifremaddr{$iface}";
		}
		elsif ($iface =~ /^atm/) {
			print "$iface PNPIF\n" if $debug;
			$cmd .= " pnpif";
			$ifispnp{$iface} = 1;
			$ifpnpremote{$iface} = "0.0.0.0";
		}
	}
	elsif (!$isloop) {
		$need_mrouted = 1;
		print "$iface is a BCAST interface\n" if $debug;
	}
	if ($iflname{$iface}) {
	  $cmd.=" lname $iflname{$iface}";
	}
	$cmd .= " $fduplex{$iface} $iflink{$iface} up";
	print "command '$cmd'\n" if $debug;
	if ($ifcmd{$iface} ne $cmd) {
	    $ifcmd{$iface}= $cmd;
	    system $cmd;
	    sleep 1;
	}

	if ($ismcast && $need_mcast_route && !$isloop) {
		$need_mcast_route = 0;
		system "route add 224.0.0.0 -netmask 0xf0000000 -interface $iflcladdr{$1}";
	}

        system("/usr/local/config/bin/policy /etc/policy.conf");
}

# IFMP packets are only sent if the interface is marked up.
# The goal here is to configure an interface we have configuration
# information on when we first encounter the interface.  We must
# also mark the interface up as we configure the address.
# With IFMP it is assumed the RUNNING flag is an indication
# of the IFMP adjacency being established.
# Interfaces that we have configuration information on are
# only ever configured and marked up the first time they
# are reported.
#
# For P2P Ipsilon interfaces (defined by interface name starting with 'ips')
# we configure only the local side and mark the interface up.
# This causes the IFMP adjacency process to start.
# Note in this case no interface route is installed, we rely on
# 'ipsrd' to pick up the interface when fully configured and install
# an appropriate route.

sub scanifs {
	local($isptp,$ismcast,$pnpremote,$isloop);
	$SIG{'HUP'}='IGNORE';
	open(IFCONFIG, "/sbin/ifconfig -a |") || return;
	while (<IFCONFIG>) {
		chop;
		# Scan through each interface line looking for interface
		# names and flags
		if (/^(.*):.*flags=.*<(.*)>/) {
			$isptp = 0;
			$ismcast = 0;
			$isloop = 0;
			@ifflags = split(/,/, $2);
			foreach $i (0 .. $#ifflags) {
				if ($ifflags[$i] eq "POINTOPOINT") {
					$isptp = 1;
				}
				if ($ifflags[$i] eq "MULTICAST") {
					$ismcast = 1;
				}
				if ($ifflags[$i] eq "LOOPBACK") {
					$isloop = 1;
				}
			}
			if (!$ifseen{$1}) {
				print "new if found: $1\n" if $debug;
				$ifseen{$1} = 1;
				# We have not seen this interface
			        # try to configure it and mark up
				&configif($1, $isptp, $ismcast, $isloop);
			} else {
				if (! $ifispnp{$iface}) {
					next;
				}
				$pnpremote = &get_remote($iface);
				if ($pnpremote ne $ifpnpremote{$iface}) {
					# remote side of a pnpif has changed
					$ifpnpremote{$iface} = $pnpremote;
					$need_mrouted = 1;
				}
			}
		}
	}
	close(IFCONFIG);
	$SIG{'HUP'} = 'handler';
}

# ifconfig <if> <addr>[/<masklen>] -?link1 -?link2 [-p2p <addr>] [lname foo]

$key = '^\s*ifconfig\s+';
$ipa = '\d+\.\d+\.\d+\.\d+';
$ifn = '[\w-]+';
$opt = '\s+-?link2';
$dup = '\s+-?link1';

sub parse_addrconf {
	local($linenum,$err);
	# Clear out any old state we might have
	undef %iflcladdr;
	undef %ifmsklen;
	undef %fduplex;
	undef %iflink;
	undef %iflname;
	undef %ifremaddr;
	undef %ifseen;
	undef %ifispnp;
	undef %ifpnpremote;
	open(ADDRCONF, $path_addrconf) || die "Can't open $path_addrconf\n";
	printf "parse '$path_addrconf'\n" if $debug;
	$linenum = 0;
	while (<ADDRCONF>) {
		chop;
		$linenum++;
		if (/^#/ || /^s*$/) {
			next;
		}
		if (!/$key($ifn)\s+($ipa)((\/)(\d+))?($dup)?($opt)(\s+-p2p\s+($ipa))?(\slname(.*))?/) {
			print "Error on line #$linenum: '$_'\n";
			$err++;
			next;
		}
		$iflcladdr{$1} = $2;
		$ifmsklen{$1} = $5;
		$fduplex{$1} = $6;
		$iflink{$1} = $7;
		$ifremaddr{$1} = $9;
		$iflname{$1} = $11;
		$ifseen{$1} = 0;
		if ($debug) {
			print "if: $1 addr: $iflcladdr{$1}";
			if ($ifmsklen{$1}) {
				printf "/$ifmsklen{$1}";
			}
			if ($ifremaddr{$1}) {
				print " -> $ifremaddr{$1}";
			}
			print "\n";
		}
	}
	close(ADDRCONF);

	if ($err) {
		print "$err error(s) encountered in parsing $path_addrconf\n";
		die;
	}
}

sub usage {
	die "Usage: pnpif [-f addrconf] [-i interval] [-d level] [-h]\n";
}

$debug = 0;
$interval = 10;
$need_mcast_route = 1;
$need_mrouted = 0;
$path_addrconf = "/etc/if.conf";
$path_pidfile = "/var/run/pnpif.pid";

while ($ARGV[0] =~ /^-/) {
	$_ = shift;
	if (/^-f(.*)/) {
		$path_addrconf = ($1 ? $1 : shift);
	}
	elsif (/^-i(.*)/) {
		$interval = ($1 ? $1 : shift);
	}
	elsif (/^-d(.*)/) {
		$debug = ($1 ? $1 : shift);
	}
	elsif (/^-h/) {
		&usage;
	}
	else {
		&usage;
	}
}

$SIG{'HUP'} = 'handler';

open(PIDFILE, "> $path_pidfile");
print PIDFILE "$$\n";
close(PIDFILE);

# We need to make sure Ipsrd or mrouted is not running when we first start
# up or reconfigure

&stop_ipsrd;

&handler;  #  Force a pass through to get the initial values

sleep 10;   #  Brief delay to allow local stuff to settle

while (1) {
	$SIG{'HUP'} = 'handler';
	&scanifs;
	if ($need_mrouted && ! $need_mcast_route) {
		&start_mrouted;
	}
	if (!$need_mcast_route) {
		&start_ipsrd;
	}
	sleep $interval;
}
